// VERSION 1.1ss
// 4/5/07

//`default_nettype none

  
/**
 * The memory module for the superscalar processor. The memory and devices are
 * all "dual-ported." Operations using the "1" ports always happen before
 * operations using the "2" ports.
 */
module memory_with_devices_ss(
    /*Memory interface*/
    IMEM_ADDR_1,
    IMEM_OUT_1,
    IMEM_ADDR_2,
    IMEM_OUT_2,
    DMEM_WE_1,
    DMEM_ADDR_1,
    DMEM_IN_1,
    DMEM_OUT_1,
    DMEM_WE_2,
    DMEM_ADDR_2,
    DMEM_IN_2,
    DMEM_OUT_2,
    /* clocks for our various io devices */
    sys_clk,
    proc_clk,
    pixel_clk,
    ps2_clk,
    exp_clk,
    GLOBAL_RST,
    GLOBAL_WE,
    /* debug signals */
    debug_mode_select,
    debug_pc_out,
    debug_imem_out, 
    debug_reg_data_in, 
    debug_reg_we, 
    debug_dmem_addr,
    debug_dmem_out, 
    debug_dmem_in, 
    debug_dmem_we,
    // Everything below is an I/O pin connection
    /*VGA ports*/
    VGA_OUT_PIXEL_CLOCK,
    VGA_COMP_SYNCH,
    VGA_OUT_BLANK_Z,
    VGA_HSYNCH,
    VGA_VSYNCH,
    VGA_OUT_RED,
    VGA_OUT_GREEN,
    VGA_OUT_BLUE,
    /*Expansion board ports*/
    SWIN,
    BTNIN,
    COUT,
    AOUT,
    LDOUT,
    /*PS2 port*/
    PS2KEYBOARD_DATA,
    PS2KEYBOARD_CLK);

   //Memory interface:
   input [15:0]   IMEM_ADDR_1, IMEM_ADDR_2;
   input [15:0]   DMEM_ADDR_1, DMEM_ADDR_2;
   input          DMEM_WE_1,   DMEM_WE_2;
   input [15:0]   DMEM_IN_1,   DMEM_IN_2;
   output [15:0]  IMEM_OUT_1,  IMEM_OUT_2;
   output [15:0]  DMEM_OUT_1,  DMEM_OUT_2;
   //Clocks/Reset
   input          sys_clk;      // general system clock - used for most of memory
   input          proc_clk;     // processor clock
   input          pixel_clk;    // drives video display
   input          ps2_clk;      // drives PS/2 keyboard driver
   input          exp_clk;      // drives expansion board 7-segs
   input          GLOBAL_RST;   // global reset
   input          GLOBAL_WE;    // global we
   //VGA:
   output         VGA_OUT_PIXEL_CLOCK;     // pixel clock for the video DAC
   output         VGA_COMP_SYNCH;          // composite sync for the video DAC
   output         VGA_OUT_BLANK_Z;         // composite blanking for the video DAC
   output         VGA_HSYNCH;              // horizontal sync for the VGA output connector
   output         VGA_VSYNCH;              // vertical sync for the VGA output connector
   output [7:0]   VGA_OUT_RED;             // RED DAC data
   output [7:0]   VGA_OUT_GREEN;           // GREEN DAC data
   output [7:0]   VGA_OUT_BLUE;            // BLUE DAC data
   //Expansion board:
   input [7:0]    SWIN;       //switches
   input [4:0]    BTNIN;      //buttons
   output [6:0]   COUT;       //7-segment display cathodes
   output [3:0]   AOUT;       //7-segment display anodes
   output [7:0]   LDOUT;      //LEDs
   //PS/2 Keyboard:
   inout          PS2KEYBOARD_DATA;
   inout          PS2KEYBOARD_CLK;
   //Debugging:
   input [2:0] 	  debug_mode_select;
   input          debug_reg_we, debug_dmem_we;
   input [15:0]   debug_pc_out, debug_imem_out, debug_reg_data_in;
   input [15:0]   debug_dmem_addr, debug_dmem_out, debug_dmem_in;

   wire [15:0]   idmem_d_out_1, idmem_d_out_2;
   wire [13:0]   vga_addr;
   wire [15:0]   vga_data;
   wire [15:0]   dio4_read_data;
   wire [10:0]   enabled_1, enabled_2;
   wire [15:0]   device_d_out_1, device_d_out_2;
   wire [15:0]   ps2_read_data;
   wire [15:0]   timer_read_data;
   //Wires added for dual-ported devices:
   wire            sel_d_out_1;
   wire            sel_d_out_2;
   wire            pipe_gwe_1, pipe_gwe_2;
   wire [15:0]     dmem_in_ss;
   wire [10:0]     enabled_ss;
   wire            dmem_we_ss;
   wire [15:0]     ps2_read_data_1,   ps2_read_data_2,   pipe_ps2_read_data_2;
   wire [15:0]     dio4_read_data_1,  dio4_read_data_2,  pipe_dio4_read_data_2;
   wire [15:0]     timer_read_data_1, timer_read_data_2, pipe_timer_read_data_2;

   //For the "double-pumping" logic, we want to use a "gwe" signal that's high
   //in the FIRST (not the last) cycle of the four-cycle "segment" before the
   //registers are written.
   wire         new_GWE;
   //We need the new_GWE signal to reset to high/enabled so that it is active
   //on the very first clock cycle that the global reset signal goes low.
   one_pulse_rst_high new_GWE_pulser (
        .clk(sys_clk),
        .rst(GLOBAL_RST),
        .btn(GLOBAL_WE),
        .pulse_out(new_GWE));
   
   //The dual-ported memory module. Please see bram_ss.v for a detailed
   //description of how to use it.
   bram_ss bram_ss_inst(
        .rst(GLOBAL_RST),
        .gwe(GLOBAL_WE),
        .addr_i1(IMEM_ADDR_1),
        .addr_i2(IMEM_ADDR_2),
        .addr_d1(DMEM_ADDR_1),
        .addr_d2(DMEM_ADDR_2),
        .addr_v({2'b11, vga_addr}),
        .clk_id(sys_clk),
        .clk_v(pixel_clk),
        .we_d1(DMEM_WE_1),
        .we_d2(DMEM_WE_2),
        .din_d1(DMEM_IN_1),
        .din_d2(DMEM_IN_2),
        .dout_i1(IMEM_OUT_1),
        .dout_i2(IMEM_OUT_2),
        .dout_d1(idmem_d_out_1),
        .dout_d2(idmem_d_out_2),
        .dout_v(vga_data));

   //vga_controller handles the VGA signals.
   vga_controller vga_cntrl_inst(
        .PIXEL_CLK(~pixel_clk),
        .RESET(GLOBAL_RST),
        .VGA_OUT_PIXEL_CLOCK(VGA_OUT_PIXEL_CLOCK),
        .VGA_COMP_SYNCH(VGA_COMP_SYNCH),
        .VGA_OUT_BLANK_Z(VGA_OUT_BLANK_Z),
        .VGA_HSYNCH(VGA_HSYNCH),
        .VGA_VSYNCH(VGA_VSYNCH),
        .VGA_OUT_RED(VGA_OUT_RED),
        .VGA_OUT_GREEN(VGA_OUT_GREEN),
        .VGA_OUT_BLUE(VGA_OUT_BLUE),
        .VGA_ADDR(vga_addr),
        .VGA_DATA(vga_data[14:0]));

   //Select DMEM_OUT from idmem or from io, depending on most-significant
   //two bits if DMEM_ADDR. If DMEM_ADDR[15:14] is 11, then select device_d_out;
   //otherwise, use idmem_d_out.
   and (sel_d_out_1, DMEM_ADDR_1[15], DMEM_ADDR_1[14]);
   mux_Nbit_2to1 #(16) mux_d_out_1 (
        .sel(sel_d_out_1), 
        .a(idmem_d_out_1), 
        .b(device_d_out_1), 
        .out(DMEM_OUT_1));
   and (sel_d_out_2, DMEM_ADDR_2[15], DMEM_ADDR_2[14]);
   mux_Nbit_2to1 #(16) mux_d_out_2 (
        .sel(sel_d_out_2), 
        .a(idmem_d_out_2), 
        .b(device_d_out_2), 
        .out(DMEM_OUT_2));
   
   //MEMORY-MAPPED I/O:
   //C000-FDFF: video memory
   //  (really, only C000-FBFF is used, but reserve up to FDFF for
   //   compatibility with P37X simulator)
   //FE00-FFFF: other I/O memory
   //FE00: keyboard status register
   //FE02: keyboard data register
   //FE03: expansion seven-segment display
   //FE04: expansion board LEDs
   //FE05: expansion board switches/buttons
   //FE06: user LEDs (D7-10)
   //FE07: user switches (SW7)/buttons (SW2-6)
   //FE08: timer status register
   //FE0A: timer interval register

   //Select which I/O device to read/write from, depending on DMEM_ADDR
   assign enabled_1 =
        DMEM_ADDR_1 == 16'hFE00 ? 11'b00000000001 : // keyboard status register (READ ONLY)
        DMEM_ADDR_1 == 16'hFE02 ? 11'b00000000100 : // keyboard data register (READ ONLY)
        DMEM_ADDR_1 == 16'hFE03 ? 11'b00000001000 : // extension board LEDs
        DMEM_ADDR_1 == 16'hFE04 ? 11'b00000010000 : // extension board Seven Segment Cathodes
        DMEM_ADDR_1 == 16'hFE05 ? 11'b00000100000 : // extension board buttons and switches (READ ONLY)
        DMEM_ADDR_1 == 16'hFE06 ? 11'b00001000000 : // currently no connected device
        DMEM_ADDR_1 == 16'hFE07 ? 11'b00010000000 : // currently no connected device
        DMEM_ADDR_1 == 16'hFE08 ? 11'b10000000000 : // timer (status) register
        DMEM_ADDR_1 == 16'hFE0A ? 11'b00100000000 : // timer interval register
                                  11'b00000000000 ; 

   assign enabled_2 =
        DMEM_ADDR_2 == 16'hFE00 ? 11'b00000000001 : // keyboard status register (READ ONLY)
        DMEM_ADDR_2 == 16'hFE02 ? 11'b00000000100 : // keyboard data register (READ ONLY)
        DMEM_ADDR_2 == 16'hFE03 ? 11'b00000001000 : // extension board LEDs
        DMEM_ADDR_2 == 16'hFE04 ? 11'b00000010000 : // extension board Seven Segment Cathodes
        DMEM_ADDR_2 == 16'hFE05 ? 11'b00000100000 : // extension board buttons and switches (READ ONLY)
        DMEM_ADDR_2 == 16'hFE06 ? 11'b00001000000 : // currently no connected device
        DMEM_ADDR_2 == 16'hFE07 ? 11'b00010000000 : // currently no connected device
        DMEM_ADDR_2 == 16'hFE08 ? 11'b10000000000 : // timer (status) register
        DMEM_ADDR_2 == 16'hFE0A ? 11'b00100000000 : // timer interval register
                                  11'b00000000000 ;  

   //Select which I/O device to read/write from, depending on DMEM_ADDR
   assign device_d_out_1 =
        DMEM_ADDR_1 == 16'hFE00 ? ps2_read_data_1   :   // keyboard status register (READ ONLY)
        DMEM_ADDR_1 == 16'hFE02 ? ps2_read_data_1   :   // keyboard data register (READ ONLY)
        DMEM_ADDR_1 == 16'hFE03 ? dio4_read_data_1  :   // extension board LEDs
        DMEM_ADDR_1 == 16'hFE04 ? dio4_read_data_1  :   // extension board Seven Segment Cathodes
        DMEM_ADDR_1 == 16'hFE05 ? dio4_read_data_1  :   // extension board buttons and switches (READ ONLY)
        DMEM_ADDR_1 == 16'hFE06 ? 16'h0000          :   // currently no connected device
        DMEM_ADDR_1 == 16'hFE07 ? 16'h0000          :   // currently no connected device
        DMEM_ADDR_1 == 16'hFE08 ? timer_read_data_1 :   // timer (status) register
        DMEM_ADDR_1 == 16'hFE0A ? timer_read_data_1 :   // timer interval register
                                  idmem_d_out_1;

   assign device_d_out_2 =
        DMEM_ADDR_2 == 16'hFE00 ? ps2_read_data_2   :   // keyboard status register (READ ONLY)
        DMEM_ADDR_2 == 16'hFE02 ? ps2_read_data_2   :   // keyboard data register (READ ONLY)
        DMEM_ADDR_2 == 16'hFE03 ? dio4_read_data_2  :   // extension board LEDs
        DMEM_ADDR_2 == 16'hFE04 ? dio4_read_data_2  :   // extension board Seven Segment Cathodes
        DMEM_ADDR_2 == 16'hFE05 ? dio4_read_data_2  :   // extension board buttons and switches (READ ONLY)
        DMEM_ADDR_2 == 16'hFE06 ? 16'h0000          :   // currently no connected device
        DMEM_ADDR_2 == 16'hFE07 ? 16'h0000          :   // currently no connected device
        DMEM_ADDR_2 == 16'hFE08 ? timer_read_data_2 :   // timer (status) register
        DMEM_ADDR_2 == 16'hFE0A ? timer_read_data_2 :   // timer interval register
                                  idmem_d_out_2;  
    
    /**
     * SUPERSCALAR:
     * The logic above for enabled_ and device_d_out_ does not need to be
     * modified, only duplicated as above. Then, all that we have to do is make
     * sure the _1 signals are used first, and the _2 signals are used next.
     * This is done by manipulating the clock and global write-enable signals
     * that are input into the modules, and muxing in the appropriate values
     * at the appropriate time.
     * 
     * We "pipeline" the global write-enable signal to give us enough state
     * information to get the device timing right. The gwe signal is high for
     * one cycle out of every four clock cycles. In the first cycle of a
     * four-cycle "section," the gwe is high and the address inputs to the
     * devices change. In the second cycle, the port 1 operations are performed;
     * however, the results of these operations are not immediately ouput, but
     * are pipelined for one cycle. In the third cycle, the port 2 operations
     * are performed, and the results for both the port 1 and port 2 operations
     * are now output. Nothing happens in the fourth cycle, and the outputs
     * remain stable.
     */
    Nbit_reg #(1,0) reg_gwe_1(
        .in(new_GWE),
        .out(pipe_gwe_1),
        .clk(sys_clk),
        .we(1'b1),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(1,0) reg_gwe_2(
        .in(pipe_gwe_1),
        .out(pipe_gwe_2),
        .clk(sys_clk),
        .we(1'b1),
        .gwe(1'b1),
        .rst(1'b0));

    /**
     * Use the pipelined gwe signal to determine which values, for port "1" or
     * port "2," to use for the devices. There are not really two sets of ports
     * on the devices; we just use the pipelined gwe signal to mux in the proper
     * inputs in two consecutive cycles, effectively "double-pumping" the
     * devices during the four-cycle gwe "section." The values below are only
     * used in the second and third cycles of the four-cycle gwe section, so
     * use the pipelined gwe signal to enable the devices when we actually want
     * to read from them.
     */
    //Wires that need to be switched into the devices:
    //DMEM_IN and DMEM_WE come in from the two datapaths in the superscalar
    //pipeline; enabled is assigned above, using DMEM_ADDR from the two
    //datapaths.
    assign dmem_in_ss       = (pipe_gwe_1) ? DMEM_IN_1 : DMEM_IN_2;
    assign enabled_ss       = (pipe_gwe_1) ? enabled_1 : enabled_2;
    assign dmem_we_ss       = (pipe_gwe_1) ? DMEM_WE_1 : DMEM_WE_2;

    //Wires that need to be switched/stored coming out of the devices:
    //ps2_read_data, dio4_read_data, and timer_read_data are connected to the
    //devices, so we assign the _1 and _2 wires from them, based on which
    //"port" is active.
    //For the "1" ports, output (and latch) the value read from the devices at
    //the end of the cycle they were read.
    Nbit_reg #(16,0) reg_ps2_read_data_1(
        .in(ps2_read_data),
        .out(ps2_read_data_1),
        .clk(sys_clk), //(sys_clk is the same as proc_clk, used for device regs)
        .we(pipe_gwe_1),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(16,0) reg_dio4_read_data_1(
        .in(dio4_read_data),
        .out(dio4_read_data_1),
        .clk(sys_clk),
        .we(pipe_gwe_1),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(16,0) reg_timer_read_data_1(
        .in(timer_read_data),
        .out(timer_read_data_1),
        .clk(sys_clk),
        .we(pipe_gwe_1),
        .gwe(1'b1),
        .rst(1'b0));
    //For the "2" ports, output the value read from the devices as soon as it is
    //read (the "true" choice in the conditional statements); this is the same
    //time that the values for the 2 ports are latched and output, so they are
    //synchronized. At the end of the cycle, latch the values so they remain
    //stable.
    Nbit_reg #(16,0) reg_ps2_read_data_2(
        .in(ps2_read_data),
        .out(pipe_ps2_read_data_2),
        .clk(sys_clk),
        .we(pipe_gwe_2),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(16,0) reg_dio4_read_data_2(
        .in(dio4_read_data),
        .out(pipe_dio4_read_data_2),
        .clk(sys_clk),
        .we(pipe_gwe_2),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(16,0) reg_timer_read_data_2(
        .in(timer_read_data),
        .out(pipe_timer_read_data_2),
        .clk(sys_clk),
        .we(pipe_gwe_2),
        .gwe(1'b1),
        .rst(1'b0));
    assign ps2_read_data_2      = (pipe_gwe_2) ? ps2_read_data :
                                                 pipe_ps2_read_data_2;
    assign dio4_read_data_2     = (pipe_gwe_2) ? dio4_read_data :
                                                 pipe_dio4_read_data_2;
    assign timer_read_data_2    = (pipe_gwe_2) ? timer_read_data :
                                                 pipe_timer_read_data_2;
    //Finally, only enable the devices on cycles when pipe_gwe_1 and pipe_gwe_2
    //are high by replacing the old GLOBAL_WE with "pipe_gwe_1 || pipe_gwe_2"
    //for the dio4 and timer modules below. We do not store to any registers in 
    //the ps2 module, so it does not use the pipe_gwe signals.

   //The expansion board controller:
   di04 dio4_inst(
        .SEV_SEG_CLK(exp_clk),            //I [0]    32 MHz clock input
        .SWIN(SWIN),                      //I [7:0]  toggle switch inputs
        .BTNIN(BTNIN),                    //I [4:0]  push-button inputs
        .COUT(COUT),                      //O [6:0]  seven-segment cathode pins
        .AOUT(AOUT),                      //O [3:0]  seven-segment anode pins
        .LDOUT(LDOUT),                    //O [7:0]  LED pins
        .reg_clk(sys_clk),
        .rst(GLOBAL_RST),
        .we(dmem_we_ss),
        .gwe(pipe_gwe_1 || pipe_gwe_2),
        .lights_data(dmem_in_ss),
        .data_select(enabled_ss[5:3]),
        .read_data(dio4_read_data),
        /* debug signals */
        .debug_mode_select(debug_mode_select),
        .debug_pc_out(debug_pc_out),
        .debug_imem_out(debug_imem_out), 
        .debug_reg_data_in(debug_reg_data_in), 
        .debug_reg_we(debug_reg_we), 
        .debug_dmem_addr(debug_dmem_addr),
        .debug_dmem_out(debug_dmem_out), 
        .debug_dmem_in(debug_dmem_in), 
        .debug_dmem_we(debug_dmem_we));
    
   //The PS/2 keyboard controller
   // no write-enable signals, as the keyboard is a read-only device
   ps2_kbd ps2_kbd_inst(
        .select(enabled_ss[2]), // chooses between KBSR (0) and KBDR (1)
        .out(ps2_read_data),    // output: KBSR or KBDR
        .reset(GLOBAL_RST),     // global reset
        // I/O pin connections
        .SYSTEM_CLOCK(sys_clk),
        .CLOCK_2MHz(ps2_clk),
        .PS2KEYBOARD_DATA(PS2KEYBOARD_DATA),
        .PS2KEYBOARD_CLK(PS2KEYBOARD_CLK));

   //The timer device
   //timer_ss timer_inst(
   timer_device timer_inst(
        .selectStatus( enabled_ss[10] ),
        .selectInterval( enabled_ss[8] ),
        .in( dmem_in_ss ),            //value to write to interval register
        .out( timer_read_data ),      //output of interval or status register
        .we( dmem_we_ss ),            //whether we're reading or writing the TMI
        .gwe(pipe_gwe_1 || pipe_gwe_2),
        .reset( GLOBAL_RST ),         //global reset
        // I/O pin connections
        .SYSTEM_CLOCK( proc_clk ));
   
endmodule
